from dataclasses import dataclass
from .....services.locale.time_pattern import TimePattern

@dataclass
class SeasonedPrice:
    validity: TimePattern
    value: float

    def __post_init__(self):
        assert self.value >= 0


from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from .hour_price import HourPrice
from .datetime_model import DateTimeModel


class AllowanceType(Enum):
    StoppedUntil = "Stopped until "
    AllowedUntil = "Allowed until "
    StoppedUntilTomorrow = "Stopped until tomorrow at "
    AllowedUntilTomorrow = "Allowed until tomorrow at "
    StoppedUntilFurtherNotice = "Stopped until further notice."
    AllowedUntilFurtherNotice = "Allowed until further notice."


@dataclass
class AllowanceObj:
    prefix_type: AllowanceType
    display_name: str = field(init=False)
    hour: int = -1
    quarter: int = -1
    datum: date = date.today()

    def __post_init__(self):
        # set display name
        match self.prefix_type:
            case AllowanceType.StoppedUntil | AllowanceType.AllowedUntil:
                self.display_name = f"{self.prefix_type.value}{self.__set_num_value(self.hour)}:{self.__set_num_value(self.quarter * 15)}."
            case AllowanceType.StoppedUntilTomorrow | AllowanceType.AllowedUntilTomorrow:
                self.display_name = (
                    f"{self.prefix_type.value}{self.__set_num_value(self.hour)}:00."
                )
            case AllowanceType.StoppedUntilFurtherNotice | AllowanceType.AllowedUntilFurtherNotice:
                self.display_name = self.prefix_type.value

    def __set_num_value(self, hour: int):
        _h = str(hour)
        if len(_h) == 1:
            return f"0{_h}"
        return _h


def set_allowance_obj(
    dtmodel: DateTimeModel, future_hours: list[HourPrice]
) -> AllowanceObj:
    if not len([hp.dt for hp in future_hours if hp.permittance > 0.0]):
        return AllowanceObj(AllowanceType.StoppedUntilFurtherNotice)
    if not len([hp.dt for hp in future_hours if hp.permittance == 0.0]):
        return AllowanceObj(AllowanceType.AllowedUntilFurtherNotice)
    first_start = min([hp.dt for hp in future_hours if hp.permittance > 0.0])
    _stopped = first_start != dtmodel._datetime
    if _stopped:
        if first_start.date() > dtmodel._date:
            return AllowanceObj(
                AllowanceType.StoppedUntilTomorrow, hour=first_start.hour
            )
        else:
            return AllowanceObj(
                AllowanceType.StoppedUntil,
                hour=first_start.hour,
                quarter=first_start.minute // 15,
            )
    else:
        first_stop = min([hp.dt for hp in future_hours if hp.permittance == 0.0])
        if first_stop.date() > dtmodel._date:
            return AllowanceObj(
                AllowanceType.AllowedUntilTomorrow, hour=first_stop.hour
            )
        else:
            return AllowanceObj(
                AllowanceType.AllowedUntil,
                hour=first_stop.hour,
                quarter=first_stop.minute // 15,
            )



from .models.hour_price import HourPrice
from ...models.hourselection.cautionhourtype import CautionHourType
from .models.hour_type import HourType
from statistics import mean


# def set_initial_permittance(
#     hour_prices: list[HourPrice], price_mean: float, price_stdev: float
# ) -> None:
#     for hp in hour_prices:
#         print(hp.dt, hp.price, price_mean, price_stdev)
#         if hp.hour_type == HourType.BelowMin:
#             hp.permittance = 1.0
#         elif hp.hour_type == HourType.AboveMax:
#             hp.permittance = 0.0
#         elif hp.price < price_mean - price_stdev:
#             hp.permittance = 1.0
#         elif hp.price > price_mean + price_stdev:
#             hp.permittance = 0.0
#         else:
#             hp.permittance = round(
#                 1.0 - ((hp.price - price_mean + price_stdev) / (2 * price_stdev)), 2
#             )
#     # return hour_prices


def set_initial_permittance(
    hours: list[HourPrice],
    price_mean: float,
    price_stdev: float,
    avg7: float | None = None,
) -> None:
    scaling_factor = scale_permittance(price_stdev)
    for hour in hours:
        if hour.hour_type == HourType.BelowMin:
            hour.permittance = 1.0
        elif hour.hour_type == HourType.AboveMax:
            hour.permittance = 0.0
        else:
            diff_avg_price = price_mean - hour.price
            diff_avg_7day_price = (avg7 or price_mean) - hour.price
            hour.permittance = mean(
                [
                    max(0, min(1, diff_avg_price / (price_mean * scaling_factor))),
                    max(
                        0,
                        min(
                            1,
                            diff_avg_7day_price
                            / ((avg7 or price_mean) * scaling_factor),
                        ),
                    ),
                ]
            )


def scale_permittance(stdev: float) -> float:
    if stdev == 0:
        return 1.0
    return max(0.1, min(1.0, 1.0 / (1.0 + stdev)))


def set_scooped_permittance(
    hour_prices: list[HourPrice], caution_hour_type: CautionHourType
) -> None:
    lo_cutoff = 0.4
    hi_cutoff = 0.75
    max_hours = 24  # todo: add support for 96 if quarterly
    match caution_hour_type:
        case CautionHourType.SUAVE:
            hi_cutoff = 0.7
        case CautionHourType.INTERMEDIATE:
            lo_cutoff = 0.5
            hi_cutoff = 0.7
        case CautionHourType.AGGRESSIVE:
            lo_cutoff = 0.6
        case CautionHourType.SCROOGE:
            lo_cutoff = 0.6
            max_hours = 8  # todo: add support for 32 if quarterly

    for hp in hour_prices:
        _t = hp.permittance
        if hp.permittance <= lo_cutoff:
            hp.permittance = 0.0
        elif hp.permittance >= hi_cutoff:
            hp.permittance = 1.0
        else:
            hp.permittance = round(hp.permittance, 2)
    # return hour_prices



from .locale_price import LocalePrice
from datetime import datetime, timedelta

#mock
from peaqevcore.services.locale.countries.sweden import SE_Gothenburg
AVG_PRICE_SEVEN_DAYS = 2.59
PRICES = [1.715, 1.665, 1.138, 1.112, 1.019, 0.925, 0.925, 0.937, 1.499, 1.7, 1.708, 1.678, 1.339, 0.925, 0.593, 0.485, 0.416, 0.412, 0.412, 0.41, 0.404, 0.409, 0.362, 0.306, 0.282, 0.198] #0 is current hour.
#mock

class PriceOverride:
    """When to override peaks"""
    peak_peak = 10

    def __init__(self, price:LocalePrice) -> None:
        self._price_type = price.price_type
        self._cost = price.value

    #if tiered price we should obv allow up to just below the next level.

    @property
    def increase(self) -> float:
        prices = PRICES #get from nordpool-data
        avg = AVG_PRICE_SEVEN_DAYS #get from sensor
        if prices[0] >= (avg*0.75):
            return 0.0

        period = self._rest_of_period()
        return self._calculate_increase(period_len=period, avg_price=avg, prices=prices)

    def _calculate_increase(self, period_len, avg_price, prices:list) -> float:
        addition = 0
        price_determination = self._calc_price_determinator(
            current_price=prices[0], 
            avg_price=avg_price, 
            prognosis=prices[1::]
            )
        try:
            while (addition * self._cost)/period_len < price_determination:
                addition += 0.1
        except ZeroDivisionError:
            pass
        return round(addition,1)

    @staticmethod
    def _calc_price_determinator(current_price:float, avg_price:float, prognosis:list) -> float:
        #dont forget to handle if price is 0 or lower.
        return 5
        
    @staticmethod
    def _rest_of_period() -> int:
        today: datetime = datetime.now().date()
        next_month = today.replace(day=28) + timedelta(days=4)
        ret: datetime = next_month - timedelta(days=next_month.day)
        return ret.day-today.day
    


# i = PriceOverride(cost=SE_Gothenburg.price)
# print(i.increase)



from statistics import mean, stdev
import logging
from ...models.hourselection.cautionhourtype import CautionHourType, MAX_HOURS

_LOGGER = logging.getLogger(__name__)


def normalize_prices(prices: list) -> list:
    min_price = min(prices)
    if min_price > 0:
        return prices
    c = 0
    if min_price <= 0:
        c = abs(min_price) + 0.01
    ret = []
    for p in prices:
        pp = p + c
        divider = min_price if min_price > 0 else c
        ret.append(round(pp - divider, 3))
    return ret


def get_offset_dict(normalized_hours: list):
    ret = {}
    _prices = [p - min(normalized_hours) for p in normalized_hours]
    average_val = mean(_prices)
    for i in range(0, len(_prices)):
        # todo: fix this to accommodate quarters also
        try:
            ret[i] = round((_prices[i] / average_val) - 1, 2)
        except:
            ret[i] = 1
    return ret


def deviation_from_mean(prices: list[float]) -> dict[int, float]:
    if not len(prices):
        return {}
    avg = mean(prices)
    devi = stdev(prices)
    deviation_dict = {}
    for i, num in enumerate(prices):
        deviation = (num - avg) / devi
        if devi < 1:
            deviation *= 0.5
        deviation_dict[i] = round(deviation, 2)
    return deviation_dict


async def async_discard_excessive_hours(hours: dict):
    """There should always be at least four regular hours before absolute_top_price kicks in."""
    while len(hours) >= 20:
        to_pop = dict(sorted(hours.items(), key=lambda item: item[1]["val"]))
        hours.pop(list(to_pop.keys())[0])
    return hours
